#include "common.h"
#include <cds/misc/chase-lev-deque.h>
#include <cds_test/stress_test.h>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <memory>
#include <thread>

using namespace std;

namespace {

typedef cds_others::ChaseLevDeque Deque;
static size_t s_nDequePushPercentage = 70;
static size_t s_nDequeStealerThreadCount = 3;
static size_t s_nDequeMainPassCount = 100000000;
static size_t s_nInitialDequePushPassCount = 10000;

class ChaseLevDequeTest : public cds_test::stress_fixture {
protected:
  static Deque *deque;
  static atomic_bool pusher_done;
  static ullong *sums;
  static ullong *succ_counts;
  static ullong push_sum;
  static ullong push_count;

  static void SetUpTestCase() {
    cds_test::config const &cfg = get_config("Misc");
    GetConfig(DequePushPercentage);
    GetConfig(DequeStealerThreadCount);
    GetConfig(DequeMainPassCount);
    GetConfig(InitialDequePushPassCount);
  }

  static void StealerThread(int index) {
    while (!pusher_done.load(memory_order_acquire)) {
      while (true) {
        int res = deque->steal();
        if (res != EMPTY && res != ABORT) {
          sums[index] += res;
          succ_counts[index]++;
        } else {
          break;
        }
      }
    }
  }

  static void MainThread(int index, int push_percentage) {
    for (size_t i = 0; i < s_nInitialDequePushPassCount; i++) {
      deque->push(i + 1);
      push_sum += i + 1;
      push_count++;
    }
    for (size_t i = 0; i < s_nDequeMainPassCount; i++) {
      if (rand(100) < push_percentage) {
        int item = rand(s_nDequeMainPassCount);
        if (item == EMPTY || item == ABORT) {
          item = 1;
        }
        deque->push(item);
        push_sum += item;
        push_count++;
      } else {
        int res = deque->take();
        if (res != EMPTY) {
          sums[index] += res;
          succ_counts[index]++;
        }
      }
    }
    // Try take() until we don't can't take any more.
    while (true) {
      int res = deque->take();
      if (res != EMPTY) {
        sums[index] += res;
        succ_counts[index]++;
      } else {
        break;
      }
    }
  }
};

atomic_bool ChaseLevDequeTest::pusher_done;
ullong *ChaseLevDequeTest::sums;
ullong *ChaseLevDequeTest::succ_counts;
ullong ChaseLevDequeTest::push_count;
ullong ChaseLevDequeTest::push_sum;
Deque *ChaseLevDequeTest::deque;

TEST_F(ChaseLevDequeTest, DequePushPopTake) {
  deque = new Deque();
  push_sum = 0;
  sums = (ullong *)calloc(1, sizeof(ullong) * (s_nDequeStealerThreadCount + 1));
  succ_counts =
      (ullong *)calloc(1, sizeof(ullong) * (s_nDequeStealerThreadCount + 1));

  // Stealer threads
  std::unique_ptr<std::thread[]> threads(
      new std::thread[s_nDequeStealerThreadCount]);
  for (ullong i = 0; i < s_nDequeStealerThreadCount; i++) {
    threads[i] = std::thread(StealerThread, i);
  }

  MainThread(s_nDequeStealerThreadCount, s_nDequePushPercentage);
  pusher_done.store(true, memory_order_release);

  for (ullong i = 0; i < s_nDequeStealerThreadCount; i++) {
    threads[i].join();
  }

  // Result analysis
  ullong received_sum = 0;
  ullong overall_count = 0;
  for (ullong i = 0; i <= s_nDequeStealerThreadCount; i++) {
    received_sum += sums[i];
    overall_count += succ_counts[i];
  }
  if (overall_count != push_count || received_sum != push_sum) {
    cout << "Incorrect deque\n";
    cout << "Push sum: " << push_sum << "\n";
    cout << "Received sum:" << received_sum << "\n";
    cout << "Push count=" << push_count << "\n";
    cout << "Received count:" << overall_count << "\n";
  }

  free(sums);
  free(succ_counts);
  delete deque;
}

} // namespace
