/*
    This file is a part of libcds - Concurrent Data Structures library

    (C) Copyright Maxim Khizhinsky (libcds.dev@gmail.com) 2006-2017

    Source code repo: http://github.com/khizmax/libcds/
    Download: http://sourceforge.net/projects/libcds/files/

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
   this
      list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
   USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "../misc/common.h"
#include "../queue/queue_type.h"

#include <algorithm>
#include <type_traits>
#include <vector>

// Sequential queue push/pop test
namespace {

static size_t s_nEnqueueCount = 4000000;
// Call 's_nEnqueueStride' number of enqueue before dequeue.
static size_t s_nEnqueueStride = 10000;
static size_t s_nQueueSize = 100000;
static size_t s_nVyukovQueuePushCount = 300000;

struct old_value {
  size_t nNo;
};

template <class Value = old_value>
class sequential_queue : public cds_test::stress_fixture {
protected:
  using value_type = Value;

  template <typename Queue>
  void test(Queue &q, size_t enqueue_count = s_nEnqueueCount) {
    value_type v;
    v.nNo = 0;
    size_t push_failure = 0;
    size_t pop_sum = 0;

    while (v.nNo < enqueue_count) {
      size_t curr_push_count =
          std::min(enqueue_count - v.nNo, s_nEnqueueStride);
      for (size_t i = 0; i < curr_push_count; i++) {
        if (q.push(v))
          ++v.nNo;
        else
          ++push_failure;
      }

      value_type res;
      while (q.pop(res)) {
        pop_sum += res.nNo;
      }
    }

    if (push_failure) {
      std::cout << "Sequential queue push error count: " << push_failure
                << "\n";
    }
    size_t supposed_sum = enqueue_count * (enqueue_count - 1) / 2;
    if (pop_sum != supposed_sum) {
      std::cout << "Sequential queue pop sum: " << pop_sum
                << " != " << supposed_sum << "\n";
    }
  }

public:
  static void SetUpTestCase() {
    cds_test::config const &cfg = get_config("SequentialQueue");
    GetConfig(EnqueueCount);
    GetConfig(EnqueueStride);
    GetConfig(QueueSize);
    GetConfig(VyukovQueuePushCount);
  }
};

using simple_sequential_queue = sequential_queue<>;

#define CDSSTRESS_Sequential_Queue_F(test_fixture, type_name)                  \
  TEST_F(test_fixture, type_name) {                                            \
    typedef queue::Types<value_type>::type_name queue_type;                    \
    std::unique_ptr<queue_type> queue(new queue_type(s_nQueueSize));           \
    test(*queue, s_nVyukovQueuePushCount);                                     \
  }

#define CDSSTRESS_Sequential_VyukovQueue(test_fixture)                         \
  CDSSTRESS_Sequential_Queue_F(test_fixture, VyukovMPMCCycleQueue_dyn);        \
  CDSSTRESS_Sequential_Queue_F(test_fixture, VyukovMPMCCycleQueue_dyn_ic);

CDSSTRESS_Sequential_VyukovQueue(simple_sequential_queue);

#undef CDSSTRESS_Sequential_Queue_F
#define CDSSTRESS_Sequential_Queue_F(test_fixture, type_name)                  \
  TEST_F(test_fixture, type_name) {                                            \
    typedef queue::Types<value_type>::type_name queue_type;                    \
    std::unique_ptr<queue_type> queue(new queue_type());                       \
    test(*queue);                                                              \
  }

#define CDSSTRESS_Sequential_MSQueue(test_fixture)                             \
  CDSSTRESS_Sequential_Queue_F(test_fixture, MSQueue_HP);                      \
  CDSSTRESS_Sequential_Queue_F(test_fixture, MSQueue_DHP);

#define CDSSTRESS_Sequential_MoirQueue(test_fixture)                           \
  CDSSTRESS_Sequential_Queue_F(test_fixture, MoirQueue_HP);                    \
  CDSSTRESS_Sequential_Queue_F(test_fixture, MoirQueue_DHP);

#define CDSSTRESS_Sequential_OptimsticQueue(test_fixture)                      \
  CDSSTRESS_Sequential_Queue_F(test_fixture, OptimisticQueue_HP);              \
  CDSSTRESS_Sequential_Queue_F(test_fixture, OptimisticQueue_DHP);

#define CDSSTRESS_Sequential_BasketQueue(test_fixture)                         \
  CDSSTRESS_Sequential_Queue_F(test_fixture, BasketQueue_HP);                  \
  CDSSTRESS_Sequential_Queue_F(test_fixture, BasketQueue_DHP);

#define CDSSTRESS_Sequential_RWQueue(test_fixture)                             \
  CDSSTRESS_Sequential_Queue_F(test_fixture, RWQueue_Spin);

#define CDSSTRESS_Sequential_WeakRingBuffer(test_fixture)                      \
  CDSSTRESS_Sequential_Queue_F(test_fixture, WeakRingBuffer_dyn);

CDSSTRESS_Sequential_MSQueue(simple_sequential_queue);
CDSSTRESS_Sequential_MoirQueue(simple_sequential_queue);
CDSSTRESS_Sequential_BasketQueue(simple_sequential_queue);
CDSSTRESS_Sequential_OptimsticQueue(simple_sequential_queue);
CDSSTRESS_Sequential_RWQueue(simple_sequential_queue);
//CDSSTRESS_Sequential_WeakRingBuffer(simple_sequential_queue);

// ********************************************************************
// SegmentedQueue test

class sequential_segmented_queue
    : public sequential_queue<>,
      public ::testing::WithParamInterface<size_t> {
  typedef sequential_queue<> base_class;

protected:
  template <typename Queue> void test() {
    size_t quasi_factor = GetParam();
    Queue q(quasi_factor);
    base_class::test(q);
  }

public:
  static std::vector<size_t> get_test_parameters() {
    cds_test::config const &cfg =
        cds_test::stress_fixture::get_config("SequentialQueue");
    bool bIterative = cfg.get_bool("SegmentedQueue_Iterate", false);
    size_t quasi_factor = cfg.get_size_t("SegmentedQueue_SegmentSize", 256);

    std::vector<size_t> args;
    if (bIterative && quasi_factor > 4) {
      for (size_t qf = 4; qf <= quasi_factor; qf *= 2)
        args.push_back(qf);
    } else {
      if (quasi_factor > 2)
        args.push_back(quasi_factor);
      else
        args.push_back(2);
    }

    return args;
  }
};

#undef CDSSTRESS_Sequential_Queue_F
#define CDSSTRESS_Sequential_Queue_F(test_fixture, type_name)                  \
  TEST_P(test_fixture, type_name) {                                            \
    typedef typename queue::Types<value_type>::type_name queue_type;           \
    test<queue_type>();                                                        \
  }

#define CDSSTRESS_Sequential_SegmentedQueue(test_fixture)                      \
  CDSSTRESS_Sequential_Queue_F(test_fixture, SegmentedQueue_HP_spin);          \
  CDSSTRESS_Sequential_Queue_F(test_fixture, SegmentedQueue_HP_spin_padding);  \
  CDSSTRESS_Sequential_Queue_F(test_fixture, SegmentedQueue_DHP_spin);

CDSSTRESS_Sequential_SegmentedQueue(sequential_segmented_queue)

#ifdef CDSTEST_GTEST_INSTANTIATE_TEST_CASE_P_HAS_4TH_ARG
    static std::string
    get_test_parameter_name(testing::TestParamInfo<size_t> const &p) {
  return std::to_string(p.param);
}
INSTANTIATE_TEST_CASE_P(
    SQ, sequential_segmented_queue,
    ::testing::ValuesIn(sequential_segmented_queue::get_test_parameters()),
    get_test_parameter_name);
#else
    INSTANTIATE_TEST_CASE_P(
        SQ, sequential_segmented_queue,
        ::testing::ValuesIn(sequential_segmented_queue::get_test_parameters()));
#endif

// ********************************************************************
// SPSC WeakRingBuffer
static size_t s_nBufferSize = 1024 * 1024;
static size_t s_nPushCount = 1000000;

class sequential_weak_ring_buffer : public cds_test::stress_fixture {
protected:
  typedef size_t value_type;

  template <typename Queue> void test(Queue &m_Queue) {
    size_t const nPushCount = s_nPushCount;
    size_t m_nPushed = 0;
    size_t m_nPushFailed = 0;
    size_t m_nPopped = 0;
    size_t m_nBadValue = 0;
    size_t m_nPopFrontFailed = 0;
    size_t m_nPopEmpty = 0;

    for (size_t i = 0; i < nPushCount; ++i) {
      size_t len = rand(1024) + 64;
      void *buf = m_Queue.back(len);
      if (buf) {
        memset(buf, len % 256, len);
        m_Queue.push_back();
        m_nPushed += len;
      } else
        ++m_nPushFailed;
    }

    while (true) {
      auto buf = m_Queue.front();
      if (buf.first) {
        m_nPopped += buf.second;

        uint8_t val = static_cast<uint8_t>(buf.second % 256);
        uint8_t const *p = reinterpret_cast<uint8_t *>(buf.first);
        for (uint8_t const *pEnd = p + buf.second; p < pEnd; ++p) {
          if (*p != val) {
            ++m_nBadValue;
            break;
          }
        }

        if (!m_Queue.pop_front())
          ++m_nPopFrontFailed;
      } else {
        ++m_nPopEmpty;
        if (m_Queue.empty())
          break;
      }
    }
  }

public:
  static void SetUpTestCase() {
    cds_test::config const &cfg = get_config("SequentialRingBuffer");

    s_nBufferSize = cfg.get_size_t("BufferSize", s_nBufferSize);
    s_nPushCount = cfg.get_size_t("PushCount", s_nPushCount);

    if (s_nBufferSize < 1024 * 64)
      s_nBufferSize = 1024 * 64;
    if (s_nPushCount == 0u)
      s_nPushCount = 1024;
  }
};

#undef CDSSTRESS_Sequential_Queue_F
#define CDSSTRESS_Sequential_Queue_F(test_fixture, type_name)                  \
  TEST_F(test_fixture, type_name) {                                            \
    typedef queue::Types<value_type>::type_name queue_type;                    \
    std::unique_ptr<queue_type> queue(new queue_type(s_nBufferSize));          \
    test(*queue);                                                              \
  }

CDSSTRESS_WeakRingBuffer_void(sequential_weak_ring_buffer)

#undef CDSSTRESS_Queue_F

} // namespace
